# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui to select a slice in a volume
"""


__authors__ = [
    "H. Payno",
]

__license__ = "MIT"
__date__ = "25/02/2021"


import os
import weakref
import numpy
import logging
from typing import Union
from contextlib import AbstractContextManager
from matplotlib import image as _matplotlib_image
from silx.gui import qt
from silx.gui.plot import PlotWidget
from tomwer.io.utils.utils import get_slice_data
from tomwer.gui.reconstruction.saaxis.dimensionwidget import DimensionWidget
from tomwer.gui.visualization.dataviewer import ImageStack
from tomwer.gui.utils.vignettes import VignettesQDialog
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.gui import icons
from tomwer.gui import settings
from tomwer.io.utils.h5pyutils import DatasetReader
from tomwer.core.process.reconstruction.scores.params import ScoreMethod
from tomwer.io.utils import get_default_directory

_logger = logging.getLogger(__name__)


class VariableScorePlot(PlotWidget):
    sigMouseWheelActive = qt.Signal(object)

    MARKER_COLOR = "#ff292199"

    def __init__(self, parent):
        PlotWidget.__init__(self, parent)
        self._scores = {}
        self.setAxesDisplayed(False)
        self.setMaximumHeight(150)
        self.setMinimumHeight(100)
        # cor marker
        # self.addXMarker(
        #     x=100, legend="cor", text="cor", draggable=False, color=self.MARKER_COLOR
        # )
        # self.corMarker = self._getMarker("cor")
        # self.corMarker.setLineWidth(3)
        # Retrieve PlotWidget's plot area widget
        plotArea = self.getWidgetHandle()
        # Set plot area custom context menu
        plotArea.setContextMenuPolicy(qt.Qt.CustomContextMenu)
        plotArea.customContextMenuRequested.connect(self._contextMenu)
        self.setInteractiveMode("select", zoomOnWheel=False)
        self.setPanWithArrowKeys(False)

    def clear(self):
        super().clear()
        # cor marker
        # self.addXMarker(
        #     x=100, legend="cor", text="cor", draggable=False, color=self.MARKER_COLOR
        # )
        # self.corMarker = self._getMarker("cor")

    def wheelEvent(self, event):
        self.sigMouseWheelActive.emit(event)

    def onMouseWheel(self, xPixel, yPixel, angleInDegrees):
        pass

    def _contextMenu(self, pos):
        """Handle plot area customContextMenuRequested signal.

        :param QPoint pos: Mouse position relative to plot area
        """
        # avoir reset zoom
        pass


class VariableSelection(qt.QWidget):
    """Widget to select which cor to select"""

    sigSelectionModeChanged = qt.Signal()
    """Signal emitted when selection mode changed"""
    sigAutoFocusLocked = qt.Signal()
    sigAutoFocusUnLocked = qt.Signal()

    sigVariableValueSelected = qt.Signal(float)
    """signal emitted when the cor value is selected by the user from
    vignettes"""

    sigScoreMethodChanged = qt.Signal()
    """Signal meit when the score method (std, tv) change"""

    def __init__(self, parent=None, variable_name=None):
        if not isinstance(variable_name, str):
            raise TypeError("variable_name should be a string")
        self._variable_name = variable_name
        qt.QWidget.__init__(self, parent)
        self._scores = {}
        self._img_width = None
        self._scan = None
        self.setLayout(qt.QGridLayout())
        # score method
        self._scoreMethodLabel = qt.QLabel("score method", self)
        self.layout().addWidget(self._scoreMethodLabel, 0, 0, 1, 1)
        self._scoreMethodCB = qt.QComboBox(self)
        self.layout().addWidget(self._scoreMethodCB, 0, 1, 1, 1)
        for method in ScoreMethod:
            self._scoreMethodCB.addItem(method.value)
        # default score method is 1 / tv
        idx = self._scoreMethodCB.findText(ScoreMethod.TV_INVERSE.value)
        self._scoreMethodCB.setCurrentIndex(idx)
        # left spacer
        lspacer = qt.QWidget(self)
        lspacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(lspacer, 0, 2, 2, 1)
        # TODO: this should be inherite by a cor plot (to display variable and absloute value)
        # cor value
        self._currentVarValue = qt.QGroupBox(self)
        self._currentVarValue.setTitle("current {} value".format(self._variable_name))
        self._currentVarValue.setLayout(qt.QFormLayout())
        self.layout().addWidget(self._currentVarValue, 0, 3, 3, 2)

        # cor selection option
        self._currentcorRB = qt.QRadioButton("current value", self)
        self.layout().addWidget(self._currentcorRB, 0, 5, 1, 1)
        self._autofocusRB = qt.QRadioButton("autofocus", self)
        self._autofocusRB.setToolTip(
            "Take the {} with the best score".format(self._variable_name)
        )
        self.layout().addWidget(self._autofocusRB, 1, 5, 1, 1)
        # lock autofocus button
        self._lockAutofocusButton = PadlockButton(self)
        self._lockAutofocusButton.setToolTip(
            "If autofocus is locked then "
            "the best center of rotation "
            "will be pick automatically"
        )
        self.layout().addWidget(self._lockAutofocusButton, 1, 7, 1, 1)
        rspacer = qt.QWidget(self)
        rspacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(rspacer, 0, 8, 2, 1)

        openVignettePixmap = icons.getQPixmap("vignettes")
        openVignetteIcon = qt.QIcon(openVignettePixmap)
        self._vignetteButton = qt.QPushButton(self)
        self._vignetteButton.setFixedSize(qt.QSize(50, 50))
        self._vignetteButton.setIconSize(qt.QSize(40, 40))
        self._vignetteButton.setIcon(openVignetteIcon)
        self.layout().addWidget(self._vignetteButton, 0, 9, 2, 1)

        # radio button group
        self._buttonGrp = qt.QButtonGroup()
        self._buttonGrp.addButton(self._currentcorRB)
        self._buttonGrp.addButton(self._autofocusRB)

        # set up
        self._currentcorRB.setChecked(True)

        # connect signal / slot
        self._currentcorRB.toggled.connect(self._selectionModeChanged)
        self._autofocusRB.toggled.connect(self._selectionModeChanged)
        self._lockAutofocusButton.toggled.connect(self._lockButtonActive)
        self._vignetteButton.released.connect(self._openVignetteMode)
        self._scoreMethodCB.currentIndexChanged.connect(self._scoreMethodChanged)

        # update widget to fit set up
        self._selectionModeChanged()

    def getVarSelectionMode(self):
        if self._currentcorRB.isChecked():
            return "current value"
        elif self._autofocusRB.isChecked():
            return "autofocus"
        else:
            raise NotImplementedError("")

    def _openVignetteMode(self):
        colormap = None
        if self.parent() and hasattr(self.parent(), "getPlotWidget"):
            master_plot = self.parent().getPlotWidget()
            colormap = master_plot.getColorBarWidget().getColormap()
        dialog = VignettesQDialog(
            value_name="{} position".format(self._variable_name),
            score_name="score",
            value_format="{:.3f}",
            score_format="{:.3f}",
            colormap=colormap,
        )
        dialog.setWindowTitle("{} - vignettes".format(self.getWindowTitle()))

        # set scores
        dialog.setScores(self._scores, score_method=self.getScoreMethod())

        if dialog.exec_() == qt.QDialog.Accepted:
            cor_selected = dialog.selectedValue()
            if cor_selected is not None:
                self.sigVariableValueSelected.emit(cor_selected)
        dialog.setAttribute(qt.Qt.WA_DeleteOnClose)

    def _selectionModeChanged(self, *args, **kwargs):
        self.sigSelectionModeChanged.emit()

    def setValue(self, relative_value):
        raise NotImplementedError("Base class")

    def getAutoFocusLockButton(self):
        return self._lockAutofocusButton

    def _scoreMethodChanged(self):
        self.sigScoreMethodChanged.emit()

    def isAutoFocusLock(self):
        return self._lockAutofocusButton.isChecked()

    def isAutoFocusActive(self):
        return self._autofocusRB.isChecked()

    def hideAutoFocusButton(self):
        self._lockAutofocusButton.hide()

    def lockAutoFocus(self, lock):
        self._currentcorRB.setEnabled(not lock)
        if lock and not self._autofocusRB.isChecked():
            self._autofocusRB.setChecked(True)
        self._lockAutofocusButton.setChecked(lock)

    def _lockButtonActive(self, lock):
        self.lockAutoFocus(lock)
        if self._lockAutofocusButton.isChecked():
            self.sigAutoFocusLocked.emit()
        else:
            self.sigAutoFocusUnLocked.emit()

    def setScores(self, scores: dict):
        self._scores = scores

    def setImgWidth(self, width):
        self._img_width = width

    def getScoreMethod(self):
        return ScoreMethod.from_value(self._scoreMethodCB.currentText())

    def setScoreMethod(self, method):
        method_value = ScoreMethod.from_value(method).value
        index = self._scoreMethodCB.findText(method_value)
        self._scoreMethodCB.setCurrentIndex(index)


class SingleValueSelection(VariableSelection):
    def __init__(self, *args, **kwargs):
        VariableSelection.__init__(self, *args, **kwargs)
        self._valueLE = qt.QLineEdit(self)
        self._valueLE.setReadOnly(True)
        self._currentVarValue.layout().addRow("value:", self._valueLE)

    def setValue(self, value):
        self._valueLE.setText("{:.2f}".format(value))

    def getWindowTitle(self):
        raise NotImplementedError("Base class")


class DelaBetaSelection(SingleValueSelection):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        idx = self._scoreMethodCB.findText(ScoreMethod.TOMO_CONSISTENCY.value)
        if idx >= 0:
            self._scoreMethodCB.removeItem(idx)

    def getWindowTitle(self):
        return "sa delta/beta vignettes"


class CorSelection(VariableSelection):
    def __init__(self, *args, **kwargs):
        VariableSelection.__init__(self, *args, **kwargs)
        self._relativeVarValueLE = qt.QLineEdit(self)
        self._relativeVarValueLE.setReadOnly(True)
        self._currentVarValue.layout().addRow("relative:", self._relativeVarValueLE)
        self._absoluteVarValueLE = qt.QLineEdit(self)
        self._absoluteVarValueLE.setReadOnly(True)
        self._currentVarValue.layout().addRow(
            "absolute:",
            self._absoluteVarValueLE,
        )
        try:
            import pyopencl  # noqa F401 pylint: disable=E0401
        except Exception:
            idx = self._scoreMethodCB.findText(ScoreMethod.TOMO_CONSISTENCY.value)
            if idx >= 0:
                self._scoreMethodCB.removeItem(idx)

    def setValue(self, relative_value):
        if relative_value is None:
            self._relativeVarValueLE.clear()
        else:
            self._relativeVarValueLE.setText("{:.3f}".format(relative_value))
        if relative_value is None or self._img_width is None:
            self._absoluteVarValueLE.clear()
        else:
            absolute_value = relative_value + self._img_width / 2.0
            self._absoluteVarValueLE.setText("{:.3f}".format(absolute_value))

    def getWindowTitle(self):
        return "sa axis - vignettes"


class PainterRotationCM(AbstractContextManager):
    """
    On enter move the painter to the position and rotate it of provided angle.
    On exits rotate back then translate back to the original position
    """

    def __init__(self, painter, x: float, y: float, angle: float):
        self.painter = painter
        self.x = x
        self.y = y
        self.angle = angle

    def __enter__(self):
        self.painter.translate(self.x, self.y)
        self.painter.rotate(self.angle)
        return self.painter

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.painter.rotate(-self.angle)
        self.painter.translate(-self.x, -self.y)


class _VariableValueLabels(qt.QWidget):
    """
    Display labels for center of rotation
    """

    # TODO: could be used to define the current value displayed by underlying
    # and the one selected in red.

    _LABEL_WIDTH = 60
    """Label width with rotation"""

    MIN_WIDTH_PER_LABEL = 40
    """Label width without rotation"""

    # DEBUG OPTIONS: Deformation is not handled so you should put rotation to 0
    PLOT_TICKS = False

    PLOT_CANVAS_LIM = False

    PLOT_LABEL_RECT = False

    def __init__(self, parent=None, angle: float = 45.0):
        qt.QWidget.__init__(self, parent)
        self._var_values = tuple()
        self._right_shift = 0
        self._slider_ticks_margin = 0
        self.rotation_angle_degree = angle

    def setVarValues(self, values):
        self._var_values = numpy.array(values)

    def setRightShift(self, shift: int):
        self._right_shift = shift

    def setSliderTicksMargin(self, margin: int):
        self._slider_ticks_margin = margin

    def getVarSteps(self) -> int:
        """
        Return step for the cor values to be displayed.
        1: display all cor values
        2: display one cor per each 2 values
        :return:
        """
        if len(self._var_values) == 0:
            return 1
        width = self.width()
        # 1.8: as we are a 45 degree then / 2.0 should be good. Dividing by 1.8
        # instead will increase visibility
        n_var_values = self._var_values.size
        for i in range(1, 9):
            if ((n_var_values / i) * self.MIN_WIDTH_PER_LABEL) < width:
                return i
        return 10

    def paintEvent(self, event):
        n_var_values = len(self._var_values)
        if n_var_values == 0:
            return
        elif n_var_values < 5:
            var_indexes = numpy.arange(n_var_values)
        else:
            var_indexes = numpy.arange(0, self._var_values.size)[:: self.getVarSteps()]

        from_ = self._slider_ticks_margin
        to_ = self.width() - self._right_shift - self._slider_ticks_margin
        var_px_positions = numpy.linspace(
            from_, to_, self._var_values.size, endpoint=True
        )

        painter = qt.QPainter(self)
        # painter.translate(self.rect().topLeft())
        font = qt.QFont("Helvetica", 10)
        painter.setFont(font)

        txt_width = self._LABEL_WIDTH
        txt_height = self.height()

        for i_var in var_indexes:
            var_value = self._var_values[i_var]
            var_px_pos = var_px_positions[i_var]
            txt_value = "{:.3f}".format(var_value)
            # apply rotation to "invert" the painter rotation requested to
            # paint oblique text
            with PainterRotationCM(
                painter=painter,
                x=var_px_pos + self._slider_ticks_margin / 2.0,
                y=0,
                angle=self.rotation_angle_degree,
            ) as l_painter:
                var_rect = qt.QRect(0, 0, txt_width, txt_height)
                l_painter.drawText(var_rect, qt.Qt.AlignLeft, txt_value)

                if self.PLOT_LABEL_RECT:
                    painter.drawRect(var_rect)

        # print all ticks
        if self.PLOT_TICKS:
            painter.setPen(qt.QPen(qt.QColor(35, 234, 32)))
            tick_positions = numpy.linspace(
                from_, to_, len(self._var_values), endpoint=True
            )
            for pos in tick_positions:
                var_rect = qt.QRect(pos, 0, 2, 6)
                painter.drawRect(var_rect)

        # print canvas
        if self.PLOT_CANVAS_LIM:
            painter.setPen(qt.QPen(qt.QColor(168, 34, 32)))
            x, y = self.width() - 2, self.height() - 2
            var_rect = qt.QRect(0, 0, x, y)
            painter.drawRect(var_rect)


class ScorePlot(qt.QWidget):
    """
    Plot to display the scores
    """

    sigConfigurationChanged = qt.Signal()
    """signal emitted when the configuration change"""

    _CENTRAL_SLICE_NAME = "central slice"

    def __init__(
        self,
        parent=None,
        variable_name=None,
        dims_colors=("#ffff5a", "#62efff", "#ff5bff"),
    ):
        if not isinstance(variable_name, str):
            raise TypeError("a variable name should be provided as a string")
        self._dim_colors = dims_colors
        self._var_values = tuple()
        self._scores = {}
        self.__scan = None

        qt.QWidget.__init__(self, parent)
        # define GUI
        self.setLayout(qt.QGridLayout())
        # main plot
        self._plot = ImageStack(self, show_overview=False)
        self._plot.getPlotWidget().setYAxisInverted(settings.Y_AXIS_DOWNWARD)

        self._plot.getPlotWidget().setKeepDataAspectRatio(True)
        self._plot._sliderDockWidget.hide()
        self._plot.getPlotWidget().getColorBarWidget().hide()
        self._plot.getPlotWidget().getPositionInfoWidget().hide()
        # add save all action
        save_all_icon = icons.getQIcon("multi-document-save")
        self._saveAllSnapshot = qt.QAction(
            save_all_icon, "save all screenshot as png", self
        )
        toolbar = self._plot.getPlotWidget().toolBar()
        toolbar.addSeparator()
        toolbar.addAction(self._saveAllSnapshot)
        # hide dock widget
        self._plot._reconsInfoDockWidget.hide()
        self._plot._tableDockWidget.hide()
        self.layout().addWidget(self._plot, 0, 0, 1, 2)
        right_shift = 40  # pixel
        # cor score plot
        self._varScore = VariableScorePlot(self)
        self._varScore.setContentsMargins(15, 10, right_shift + 15, 0)
        self.layout().addWidget(self._varScore, 1, 0, 1, 2)
        # slider
        self._varSlider = _VarSlider(self, orientation=qt.Qt.Horizontal)
        self.layout().addWidget(self._varSlider, 2, 0, 1, 2)
        self._varSlider.setContentsMargins(0, 0, right_shift, 0)
        self.setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.layout().addWidget(self._varSlider, 3, 0, 1, 2)
        # cor labels
        self._varLabels = _VariableValueLabels(self)
        self._varLabels.setContentsMargins(0, 0, 0, 0)
        self._varLabels.setFixedHeight(60)
        self.layout().addWidget(self._varLabels, 4, 0, 1, 2)
        self._varLabels.setRightShift(right_shift)
        self._varLabels.setSliderTicksMargin(15)
        # cor value
        self._varValueWidget = self._CorSlectionBuilder(self, variable_name)
        self._varValueWidget.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self._varValueWidget, 5, 0, 1, 1)
        # voxel size
        self._voxelSizeW = DimensionWidget(
            title="Voxel size",
            dims_name=("x:", "y:", "z:"),
            dims_colors=self._dim_colors,
        )
        self._voxelSizeW.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self._voxelSizeW, 6, 0, 1, 1)
        # volume size W
        self._volumeSizeW = DimensionWidget(
            title="Volume size",
            dims_name=("x:", "y:", "z:"),
            dims_colors=self._dim_colors,
        )
        self._volumeSizeW.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self._volumeSizeW, 6, 0, 1, 1)
        # axis
        self.setContentsMargins(0, 0, 0, 0)
        self._axisLabel = qt.QLabel(parent=self)
        icon = icons.getQIcon("axis")
        self._axisLabel.setPixmap(icon.pixmap(qt.QSize(96, 96)))
        self.layout().addWidget(self._axisLabel, 7, 1, 2, 1)

        # connect signal / slot
        self._varSlider.valueChanged.connect(self._plot.setCurrentUrlIndex)
        self._varSlider.valueChanged.connect(self._sliderReleased)
        self._varSlider.valueChanged.connect(self._updateVarValue)
        self._varScore.sigMouseWheelActive.connect(self._propagateSPWheelEvent)
        self._varValueWidget.sigAutoFocusLocked.connect(self._autoFocusLockChanged)
        self._varValueWidget.sigAutoFocusUnLocked.connect(self._autoFocusLockChanged)
        self._varValueWidget.sigSelectionModeChanged.connect(
            self._varSelectionModeChanged
        )
        self._varValueWidget.sigVariableValueSelected.connect(
            self._varSelectedFromVignettes
        )
        self._varValueWidget.sigScoreMethodChanged.connect(self._updateScores)
        self._saveAllSnapshot.triggered.connect(self._saveAllReconstructedSlices)

        # set up
        # for now we don't want to use the volume and voxel size
        self._volumeSizeW.hide()
        self._voxelSizeW.hide()
        self._axisLabel.hide()

    def __init_subclass__(cls, constructor, **kwargs):  # pylint: disable=E0302
        super().__init_subclass__()
        cls._CorSlectionBuilder = constructor

    def getScoreMethod(self):
        return self._varValueWidget.getScoreMethod()

    def setScoreMethod(self, method):
        self._varValueWidget.setScoreMethod(method)

    def _configurationChanged(self, *args, **kwargs):
        self.sigConfigurationChanged.emit()

    def getPlotWidget(self):
        return self._plot.getPlotWidget()

    def _stopAnimationThread(self):
        self._plot._freeLoadingThreads()
        if self._plot._plot.updateThread.is_alive():
            self._plot._plot.updateThread.stop()

    def close(self):
        self._plot.close()
        self._plot = None
        super().close()

    def getCurrentVarValue(self):
        cor_idx = self._varSlider.value()
        if cor_idx < len(self._var_values):
            return self._var_values[cor_idx]

    def setCurrentVarValue(self, value):
        self._varSlider.setVarValue(value=value)
        self._varValueWidget.setValue(value)

    def lockAutoFocus(self, lock):
        self._varValueWidget.lockAutoFocus(lock)

    def isAutoFocusLock(self):
        return self._varValueWidget.isAutoFocusLock()

    def hideAutoFocusButton(self):
        self._varValueWidget.hideAutoFocusButton()

    def isAutoFocusActive(self):
        return self._varValueWidget.isAutoFocusActive()

    def _autoFocusLockChanged(self):
        if self.isAutoFocusLock():
            self._applyAutofocus()
        self._configurationChanged()

    def getVarSelectionMode(self):
        return self._varValueWidget.getVarSelectionMode()

    def _applyAutofocus(self):
        raise NotImplementedError("Base class")

    def _varSelectionModeChanged(self):
        if self.getVarSelectionMode() == "autofocus":
            self._applyAutofocus()

    def _updateVarValue(self):
        self._varValueWidget.setValue(self.getCurrentVarValue())

    def _varSelectedFromVignettes(self, value):
        self._varSlider.setVarValue(value)

    def _propagateSPWheelEvent(self, event):
        self._varSlider.wheelEvent(event)

    def _get_closest_var(self, value):
        """return the closest cor value to value or None if no cor value
        defined"""
        if len(self._var_values) > 0:
            idx_closest = numpy.argmin(numpy.abs(self._var_values - value))
            return self._var_values[idx_closest]
        else:
            return None

    def _markerReleased(self):
        return

    def _sliderReleased(self):
        return

    def setScan(self, scan):
        self.clear()
        self.__scan = weakref.ref(scan)

    def _updateScores(self):
        raise NotImplementedError("Base class")

    def setVarScores(
        self,
        scores: dict,
        score_method: Union[str, ScoreMethod],
        img_width=None,
        update_only_scores=False,
    ):
        """

        :param dict scores: cor value (float) as key and
                            tuple(url: DataUrl, score: float) as value
        """
        score_method = ScoreMethod.from_value(score_method)
        if not update_only_scores:
            self.clear()
            self._scores = scores
            self.setImgWidth(img_width)
        # set image width to deduce absolute position from relative
        self._var_values = []
        score_list = []
        self._urls = []

        if scores is None or len(scores) == 0:
            return
        for cor_value, cor_info in scores.items():
            url, score = cor_info
            if score is None:
                continue  # in case score calculation failed
            self._var_values.append(cor_value)
            score_list.append(score.get(score_method))
            self._urls.append(url)
        # insure cor and scores are numpy arrays
        self._var_values = list(self._var_values)
        score_list = numpy.array(score_list)
        # set zoom a priori
        if not update_only_scores and len(self._urls) > 0:
            try:
                with DatasetReader(url=self._urls[0]) as dataset:
                    dataset.shape
            except Exception as e:
                _logger.warning(e)
            else:
                self._plot.setResetZoomOnNextIteration(True)

        if not update_only_scores:
            self._plot.setUrls(urls=self._urls)
            self._varSlider.setVarValues(self._var_values)
            old = self._varSlider.blockSignals(True)
            if len(self._var_values) > 0:
                self._varSlider.setVarValue(self._var_values[0])
            self._varSlider.blockSignals(old)

        old = self._varValueWidget.blockSignals(True)
        self.setScoreMethod(score_method)
        self._varValueWidget.blockSignals(old)

        self._varScore.addCurve(
            x=numpy.array(self._var_values),
            y=score_list,
            baseline=0,
            fill=True,
            color="#0288d190",
        )
        self._varLabels.setVarValues(self._var_values)
        # update cor marker position according to the slider position
        self._sliderReleased()
        self._varValueWidget.setScores(scores)
        scan = self.__scan() if self.__scan else None
        if scan is not None and self.isAutoFocusActive():
            self._applyAutofocus()

    def setImgWidth(self, width):
        self._varValueWidget.setImgWidth(width)

    def clear(self):
        self._varSlider.setVarValues(tuple())
        self._varScore.clear()
        self._varLabels.setVarValues(tuple())
        self._plot.getPlotWidget().clear()
        self._varValueWidget.setImgWidth(None)

    def setVoxelSize(self, dim0, dim1, dim2, unit):
        """

        :param float dim0:
        :param float dim1:
        :param float dim2:
        :param Union[str,float] unit:
        """
        self._voxelSizeW.setValues(dim0=dim0, dim1=dim1, dim2=dim2, unit=unit)

    def setVolumeSize(self, dim0, dim1, dim2, unit):
        """

        :param float dim0:
        :param float dim1:
        :param float dim2:
        :param Union[str,float] unit:
        """
        self._volumeSizeW.setValues(dim0=dim0, dim1=dim1, dim2=dim2, unit=unit)

    def getDim0N(self):
        """return the number of elements in dimension 0 according to voxel
        size and volume size"""
        return int(
            numpy.ceil(
                self._volumeSizeW.getDim0Value()[0] / self._voxelSizeW.getDim0Value()[0]
            )
        )

    def getDim1N(self):
        """return the number of elements in dimension 1 according to voxel
        size and volume size"""
        return int(
            numpy.ceil(
                self._volumeSizeW.getDim1Value()[0] / self._voxelSizeW.getDim1Value()[0]
            )
        )

    def getDim2N(self):
        """return the number of elements in dimension 2 according to voxel
        size and volume size"""
        return int(
            numpy.ceil(
                self._volumeSizeW.getDim2Value()[0] / self._voxelSizeW.getDim2Value()[0]
            )
        )

    def _saveAllReconstructedSlices(self):  # pragma: no cover
        """
        save all reconstructed slices to a folder provided by the user
        """
        dialog = qt.QFileDialog(self, directory=get_default_directory())
        dialog.setFileMode(qt.QFileDialog.DirectoryOnly)
        if not dialog.exec_():
            dialog.close()
            return

        output_folder = dialog.selectedFiles()[0]
        try:
            self.saveReconstructedSlicesTo(output_folder=output_folder)
        except Exception as e:
            _logger.error(f"Fail to save snapshots. Error is {e}")
        else:
            _logger.info(f"snapshots have been saved to {output_folder}")

    def saveReconstructedSlicesTo(self, output_folder):
        """dump all slices with score as name"""
        # make sure the output dir exists
        os.makedirs(output_folder, exist_ok=True)

        i_missing_score = 0
        for score, slice_url in zip(self._var_values, self._urls):
            try:
                data = get_slice_data(slice_url)
                if score is None:
                    file_name = f"no_score_{i_missing_score}.png"
                    i_missing_score += 1
                else:
                    file_name = f"{score}.png"
                _matplotlib_image.imsave(
                    os.path.join(output_folder, file_name),
                    data,
                    format="png",
                )

            except Exception as e:
                _logger.error(f"Failed to dump {slice_url.path()}. Raison is {e}")


class _VarSlider(qt.QWidget):
    def __init__(self, parent, orientation):
        self._values = None
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._slider = qt.QSlider(self)
        self.layout().addWidget(self._slider)
        self._slider.setMaximum(0)
        self._slider.setOrientation(orientation)
        self._slider.setTickPosition(qt.QSlider.TicksBelow)

        # connect signal / slot
        self.valueChanged = self._slider.valueChanged

    def wheelEvent(self, event) -> None:
        self._slider.wheelEvent(event)

    def setVarValues(self, values):
        self._slider.setRange(0, len(values) - 1)
        self._values = numpy.array(values, dtype=numpy.float32)

    def setVarValue(self, value):
        if value in self._values:
            where = numpy.where(self._values == value)
            if len(where) > 0:
                value = where[0]
                if isinstance(value, numpy.ndarray):
                    value = value[0]
                self._slider.setValue(value)

    def value(self):
        return self._slider.value()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui relative frame difference display
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "23/03/2022"


import functools
import logging
import os
import numpy
from processview.core.dataset import DatasetIdentifier
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.image import shift_img
from tomwer.gui.settings import Y_AXIS_DOWNWARD
from tomwer.gui.utils.completer import UrlCompleterDialog
from tomwer.io.utils import get_slice_data
from tomwer.gui.qfolderdialog import QDataDialog
from tomwer.gui.reconstruction.axis.CompareImages import CompareImages as _CompareImages
from tomwer.gui.visualization.diffviewer.shiftwidget import TwoFramesShiftTab
from silx.gui import qt
from silx.io.url import DataUrl
from silx.utils.enum import Enum as _Enum
from silx.gui import icons as silx_icons

_logger = logging.getLogger(__name__)


class CompareImages(_CompareImages):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        # ignore Pan with arrow keys. This should be used by the "arrowControlWidget"
        self.getPlot().setPanWithArrowKeys(False)
        self._firstVisible = True

    def setVisible(self, bool):
        if self._firstVisible:
            # hack to enforce matplotlib to replot the widget.
            # because it looks like it is missing to find the correct viewport
            # when not visible
            self.getPlot().resetZoom()
            self._firstVisible = False
        super().setVisible(bool)


class _FrameSelector(qt.QWidget):
    """Selector to select a frame from dark, flat, projection (normalized or
    not) and a reconstruction"""

    sigCorrectionChanged = qt.Signal()
    """signal emitted when the correction changed"""
    sigSelectedUrlChanged = qt.Signal()
    """signal emitted when the selected url changed"""

    class FrameType(_Enum):
        DARKS = "darks"
        FLATS = "flats"
        PROJ = "projections"
        ALIGN_PROJ = "alignment projections"
        RECON_SLICES = "slices reconstructed"

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self._scan = None
        self._currentFrameUrlsText = []

        self.setLayout(qt.QGridLayout())
        self.layout().addWidget(qt.QLabel("frame type", self), 0, 0, 1, 1)
        self._frameTypeCB = qt.QComboBox(self)
        for frame_type in _FrameSelector.FrameType:
            self._frameTypeCB.addItem(frame_type.value)
        self.layout().addWidget(self._frameTypeCB, 0, 1, 1, 3)

        self.layout().addWidget(qt.QLabel("frame", self), 1, 0, 1, 1)
        self._frameUrlCB = qt.QComboBox(self)
        self.layout().addWidget(self._frameUrlCB, 1, 1, 1, 3)
        search_icon = silx_icons.getQIcon("zoom")
        self._researchUrlButton = qt.QPushButton(parent=self, icon=search_icon)
        self._researchUrlButton.setToolTip("search for a specific url")
        self._researchUrlButton.setMaximumWidth(30)
        self._researchUrlButton.setSizePolicy(
            qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed
        )
        self.layout().addWidget(self._researchUrlButton, 1, 4, 1, 1)

        self._proj_normalized = qt.QCheckBox("corrected", self)
        self.layout().addWidget(self._proj_normalized, 2, 1, 2, 1)

        # connect signal / slot
        self._frameTypeCB.currentIndexChanged.connect(self._typeChanged)
        self._frameUrlCB.currentIndexChanged.connect(self._urlChanged)
        self._proj_normalized.toggled.connect(self._correctionChanged)
        self._researchUrlButton.released.connect(self._searchUrl)

        # default settings
        self._proj_normalized.setChecked(True)
        index = self._frameTypeCB.findText(self.FrameType.PROJ.value)
        self._frameTypeCB.setCurrentIndex(index)

    def setScan(self, scan):
        self._scan = scan
        self._typeChanged()

    def getScan(self):
        return self._scan

    def getCurrentUrl(self):
        if self._frameUrlCB.currentText() != "":
            return DataUrl(path=self._frameUrlCB.currentData(qt.Qt.UserRole))
        else:
            return None

    def getTypeSelected(self):
        return self.FrameType.from_value(self._frameTypeCB.currentText())

    def _typeChanged(self, *args, **kwargs):
        type_selected = self.FrameType.from_value(self.getTypeSelected())
        self._proj_normalized.setVisible(
            type_selected in (self.FrameType.PROJ, self.FrameType.ALIGN_PROJ)
        )
        if self._scan is None:
            return

        urls = []
        self._currentFrameUrlsText.clear()
        self._frameUrlCB.clear()
        if type_selected == self.FrameType.DARKS:
            if self._scan.darks is not None:
                urls = self._scan.darks.values()
        elif type_selected == self.FrameType.FLATS:
            if self._scan.flats is not None:
                urls = self._scan.flats.values()
        elif type_selected == self.FrameType.PROJ:
            if self._scan.projections is not None:
                urls = self._scan.projections.values()
        elif type_selected == self.FrameType.ALIGN_PROJ:
            if self._scan.alignment_projections is not None:
                urls = self._scan.alignment_projections.values()
        elif type_selected == self.FrameType.RECON_SLICES:
            urls = self._scan.get_reconstructions_urls()
        else:
            raise ValueError("Type {} not managed".format(type_selected))
        urls = sorted(urls, key=lambda url: url.path())
        for url in urls:
            if url.data_slice() is not None:
                text = "?slice=".join(
                    (os.path.basename(url.file_path()), str(url.data_slice()))
                )
            else:
                text = os.path.basename(url.file_path())
            user_data = url.path()
            self._frameUrlCB.addItem(text, user_data)
            self._currentFrameUrlsText.append(text)

    def _urlChanged(self, *args, **kwargs):
        self.sigSelectedUrlChanged.emit()

    def _searchUrl(self):
        current_url = self._frameUrlCB.currentText()
        dialog = UrlCompleterDialog(
            current_url=current_url,
            urls=tuple(self._currentFrameUrlsText),
        )
        if dialog.exec():
            new_url = dialog.selected_url()
            idx = self._frameUrlCB.findText(new_url)
            if idx >= 0:
                self._frameUrlCB.setCurrentIndex(idx)
            else:
                _logger.warning(f"Unable to find {new_url} in the list of urls")

    def _correctionChanged(self, *args, **kwargs):
        self.sigCorrectionChanged.emit()

    def needToNormalize(self):
        """

        :return: True if the data need to be treated by a dark and flat field
                 normalization.
        """
        return self._proj_normalized.isChecked() and self.getTypeSelected() in (
            self.FrameType.PROJ,
            self.FrameType.ALIGN_PROJ,
        )

    def clear(self):
        self._scan = None
        self._frameUrlCB.clear()
        self._currentFrameUrlsText = []


class _FramesSelector(qt.QWidget):
    """Selector to select a frame from dark, flat, projection (normalized or
    not) and a reconstruction"""

    sigLeftFrameUpdateReq = qt.Signal()
    """signal emit when a left frame update is requested"""
    sigRightFrameUpdateReq = qt.Signal()
    """signal emit when a right frame update is requested"""

    sigLeftScanChanged = qt.Signal()
    """signal emit when the scan for the left display changed"""
    sigRightScanChanged = qt.Signal()
    """signal emit when the scan for the right display changed"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        self._leftScanCB = _ScanComboBox(self)
        self.layout().addWidget(self._leftScanCB, 0, 0, 1, 4)
        self._rightScanCB = _ScanComboBox(self)
        self.layout().addWidget(self._rightScanCB, 0, 4, 1, 4)

        self._leftSelector = _FrameSelector(parent=self)
        self.layout().addWidget(self._leftSelector, 1, 0, 3, 4)
        self._rightSelector = _FrameSelector(parent=self)
        self.layout().addWidget(self._rightSelector, 1, 4, 3, 4)

        # signal / slot connection
        self._leftSelector.sigCorrectionChanged.connect(self._leftFrameUpdReq)
        self._rightSelector.sigCorrectionChanged.connect(self._rightFrameUpdReq)
        self._leftSelector.sigSelectedUrlChanged.connect(self._leftFrameUpdReq)
        self._rightSelector.sigSelectedUrlChanged.connect(self._rightFrameUpdReq)
        self._leftScanCB.sigScanChanged.connect(self._leftScanChanged)
        self._rightScanCB.sigScanChanged.connect(self._rightScanChanged)
        self._leftScanCB.sigRequestScanAdd.connect(self._addScanFrmDialogLeft)
        self._rightScanCB.sigRequestScanAdd.connect(self._addScanFrmDialogRight)

    def getLeftScan(self):
        return self._leftSelector.getScan()

    def setLeftScan(self, scan):
        self._leftScanCB.setScan(scan.get_identifier())
        self._leftSelector.setScan(scan)

    def getRightScan(self):
        return self._rightSelector.getScan()

    def setRightScan(self, scan):
        self._rightScanCB.setScan(scan.get_identifier())
        self._rightSelector.setScan(scan)

    def addScan(self, scan):
        if not isinstance(scan, TomwerScanBase):
            raise TypeError("scan should be an instance of {}".format(TomwerScanBase))
        self._leftScanCB.addScan(scan.get_identifier())
        self._rightScanCB.addScan(scan.get_identifier())

    def _leftScanChanged(self):
        self._leftSelector.setScan(self._leftScanCB.getCurrentScan())

    def _rightScanChanged(self):
        self._rightSelector.setScan(self._rightScanCB.getCurrentScan())

    def getLeftUrl(self):
        return self._leftSelector.getCurrentUrl()

    def needToNormalizeLeft(self):
        return self._leftSelector.needToNormalize()

    def needToNormalizeRight(self):
        return self._rightSelector.needToNormalize()

    def getRightUrl(self):
        return self._rightSelector.getCurrentUrl()

    def _rightFrameUpdReq(self):
        self.sigRightFrameUpdateReq.emit()

    def _leftFrameUpdReq(self):
        self.sigLeftFrameUpdateReq.emit()

    def _addScanFrmDialogLeft(self):
        self._addScanFrmDialog(link_to="left")

    def _addScanFrmDialogRight(self):
        self._addScanFrmDialog(link_to="right")

    def _addScanFrmDialog(self, link_to: str):
        assert link_to in ("left", "right"), "add scan should be in " "'left', 'right'"

        dialog = QDataDialog(self, multiSelection=True)

        if not dialog.exec_():
            dialog.close()
            return

        foldersSelected = dialog.files_selected()
        scan_to_set = None
        for folder in foldersSelected:
            try:
                scans = ScanFactory.create_scan_objects(scan_path=folder)
            except Exception as e:
                _logger.error(
                    "cannot create scan instances from {}. Error is {}".format(
                        folder, e
                    )
                )
            else:
                for scan in scans:
                    self._leftScanCB.addScan(scan.get_identifier())
                    self._rightScanCB.addScan(scan.get_identifier())
                    scan_to_set = scan

        if scan_to_set is not None:
            if link_to == "left":
                self._leftScanCB.setScan(scan_id=scan_to_set.get_identifier())
            else:
                self._rightScanCB.setScan(scan_id=scan_to_set.get_identifier())

    def clear(self):
        self._leftScanCB.clear()
        self._rightScanCB.clear()
        self._leftSelector.clear()
        self._rightSelector.clear()


class DiffFrameViewer(qt.QMainWindow):
    """
    Widget used to compare two frames using the silx CompareImages widget.

    User can select a reconstruction, a projection (normalized or not), flat
    or dark.
    """

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        # add frame selector
        self._framesSelector = _FramesSelector(parent=self)
        self._framesSelector.setContentsMargins(0, 0, 0, 0)
        self._framesSelector.layout().setContentsMargins(0, 0, 0, 0)
        self._framesSelectorDW = qt.QDockWidget(parent=self)
        self._framesSelectorDW.setWindowTitle("inputs")
        self._framesSelectorDW.setWidget(self._framesSelector)
        self._framesSelectorDW.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.addDockWidget(qt.Qt.TopDockWidgetArea, self._framesSelectorDW)

        # add shift dock widget
        self._shiftsWidget = TwoFramesShiftTab(self)
        self._shiftsWidget.setContentsMargins(2, 2, 2, 2)

        self._shiftDW = qt.QDockWidget(parent=self)
        self._shiftDW.setWidget(self._shiftsWidget)
        self._shiftDW.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._shiftDW.setWindowTitle("shifts")
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._shiftDW)

        # define central widget
        self._mainWidget = CompareImages(parent=self)
        self._mainWidget.setVisualizationMode(
            CompareImages.VisualizationMode.COMPOSITE_A_MINUS_B
        )
        self._mainWidget.getPlot().setYAxisInverted(Y_AXIS_DOWNWARD)

        self.setCentralWidget(self._mainWidget)

        # toolbar
        # add toolbar
        toolbar = qt.QToolBar(self)
        toolbar.setMovable(False)
        toolbar.setFloatable(False)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        # set up
        self._mainWidget.setAutoResetZoom(False)
        self._shiftsWidget.setFocus(qt.Qt.OtherFocusReason)

        # connect signal / slot
        self._framesSelector.sigRightFrameUpdateReq.connect(self._resetRightFrame)
        self._framesSelector.sigLeftFrameUpdateReq.connect(self._resetLeftFrame)
        self._shiftsWidget.sigShiftsChanged.connect(self._frameShiftsChanged)

    def getShiftsWidget(self):
        return self._shiftsWidget

    def addScan(self, scan):
        self._framesSelector.addScan(scan=scan)

    def getLeftScan(self) -> TomwerScanBase:
        return self._framesSelector.getLeftScan()

    def setLeftScan(self, scan: TomwerScanBase):
        return self._framesSelector.setLeftScan(scan)

    def getRightScan(self) -> TomwerScanBase:
        return self._framesSelector.getRightScan()

    def setRightScan(self, scan: TomwerScanBase):
        return self._framesSelector.setRightScan(scan)

    def getShiftFrameA(self) -> tuple:
        return self._shiftsWidget.getFrameAShift()

    def getShiftFrameB(self) -> tuple:
        return self._shiftsWidget.getFrameBShift()

    def isFrameALRFlip(self) -> bool:
        return self._shiftsWidget.isFrameALRFlip()

    def isFrameBLRFlip(self) -> bool:
        return self._shiftsWidget.isFrameBLRFlip()

    def _resetLeftFrame(self):
        scan = self.getLeftScan()
        if scan is None:
            return
        url = self._framesSelector.getLeftUrl()
        if url is None:
            return
        assert isinstance(url, DataUrl)
        data = self._get_data(url.path())
        if self._framesSelector.needToNormalizeLeft():
            data = scan.data_flat_field_correction(
                data=data, index=scan.get_url_proj_index(url)
            )
        if data.ndim != 2:
            if data.ndim == 3 and data.shape[0] == 1:
                data = data.reshape((data.shape[1], data.shape[2]))
            else:
                _logger.error("Cannot display {}. Should be 2D".format(url.path()))
                return

        if self.isFrameALRFlip():
            data = numpy.fliplr(data)
        shifted_image = shift_img(
            data=data, dx=self.getShiftFrameA()[0], dy=self.getShiftFrameA()[1]
        )
        self._mainWidget.setImage1(shifted_image)

    @functools.lru_cache(maxsize=4)
    def _get_data(self, url: str):
        if not isinstance(url, str):
            raise TypeError(f"url should be a str not {type(url)}")
        return get_slice_data(DataUrl(path=url))

    def _resetRightFrame(self):
        scan = self.getRightScan()
        if scan is None:
            return
        url = self._framesSelector.getRightUrl()
        if url is None:
            return
        assert isinstance(url, DataUrl)
        data = self._get_data(url.path())
        if self._framesSelector.needToNormalizeRight():
            data = scan.data_flat_field_correction(
                data=data, index=scan.get_url_proj_index(url)
            )
        if data.ndim != 2:
            if data.ndim == 3 and data.shape[0] == 1:
                data = data.reshape((data.shape[1], data.shape[2]))
            else:
                _logger.error("Cannot display {}. Should be 2D".format(url.path()))
                return
        if self.isFrameBLRFlip():
            data = numpy.fliplr(data)
        shifted_image = shift_img(
            data=data, dx=self.getShiftFrameB()[0], dy=self.getShiftFrameB()[1]
        )
        self._mainWidget.setImage2(shifted_image)

    def _frameShiftsChanged(self):
        self._resetLeftFrame()
        self._resetRightFrame()


class _ScanComboBox(qt.QWidget):
    sigScanChanged = qt.Signal()
    """Signal emit when the scan change"""

    sigRequestScanAdd = qt.Signal()
    """Signal emit when the user request to add a scan from a dialog"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self._scans = set()
        self.setLayout(qt.QHBoxLayout())
        self.setContentsMargins(0, 0, 0, 0)
        self.layout().setContentsMargins(0, 0, 0, 0)
        # scan combobox
        self._scansCB = qt.QComboBox(self)
        self.layout().addWidget(self._scansCB)
        # add scan button
        self._addButton = qt.QPushButton(self)
        self._addButton.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        self._addButton.setMaximumSize(25, 25)
        # self._addButton.setFixedSize(30, 30)
        style = qt.QApplication.instance().style()
        icon = style.standardIcon(qt.QStyle.SP_DirIcon)
        self._addButton.setIcon(icon)
        self.layout().addWidget(self._addButton)

        # connect signal / slot
        self._scansCB.currentIndexChanged.connect(self._scanChanged)
        self._addButton.pressed.connect(self._addScanCallback)

    def clear(self):
        self._scans = set()
        self._scansCB.clear()

    def addScan(self, scan_id: DatasetIdentifier):
        if not isinstance(scan_id, DatasetIdentifier):
            raise TypeError("scan should be the scan identifier")
        if scan_id not in self._scans:
            self._scans.add(scan_id)
            self._scansCB.addItem(str(scan_id), scan_id)

    def setScan(self, scan_id: DatasetIdentifier):
        if not isinstance(scan_id, DatasetIdentifier):
            raise TypeError("scan should be the scan identifier")
        idx = self._scansCB.findText(str(scan_id))
        self._scansCB.setCurrentIndex(idx)

    def _scanChanged(self, *args, **kwargs):
        self.sigScanChanged.emit()

    def getCurrentScan(self) -> TomwerScanBase:
        current_idx = self._scansCB.currentIndex()
        dataset_id = self._scansCB.itemData(current_idx)
        try:
            scan = DatasetIdentifier.recreate_dataset(dataset_id)
        except Exception as e:
            _logger.error(
                "Fail to recreate dataset from {}. Reason is {}."
                "".format(str(dataset_id), e)
            )
            return None
        else:
            return scan

    def _addScanCallback(self):
        self.sigRequestScanAdd.emit()



import os
import pytest
import numpy
from tomoscan.scanbase import _FOV
from tomoscan.esrf.scan.hdf5scan import ImageKey
from tomoscan.unitsystem.metricsystem import MetricSystem
from tomoscan.unitsystem.energysystem import EnergySI
from tomwer.core.scan.hdf5scan import HDF5TomoScan
from tomwer.gui.edit.nxtomoeditor import NXtomoEditor, _TranslationMetricEntry
from nxtomomill.nexus.nxtomo import NXtomo
from tomwer.tests.conftest import qtapp  # noqa F401
from silx.gui import qt


@pytest.mark.parametrize("x_pixel_size", (None, 0.12))
@pytest.mark.parametrize("y_pixel_size", (None, 0.0065))
@pytest.mark.parametrize("field_of_view", _FOV.values())
@pytest.mark.parametrize("distance", (None, 1.2))
@pytest.mark.parametrize("energy", (None, 23.5))
@pytest.mark.parametrize("x_flipped", (True, False))
@pytest.mark.parametrize("y_flipped", (True, False))
@pytest.mark.parametrize("x_translation", (None, numpy.ones(12), numpy.arange(12)))
@pytest.mark.parametrize("z_translation", (None, numpy.zeros(12), numpy.arange(12, 24)))
def test_nx_editor(
    tmp_path,
    qtapp,  # noqa F811
    x_pixel_size,
    y_pixel_size,
    field_of_view,
    distance,
    energy,
    x_flipped,
    y_flipped,
    x_translation,
    z_translation,
):
    # 1.0 create nx tomo with raw data
    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.x_pixel_size = x_pixel_size
    nx_tomo.instrument.detector.y_pixel_size = y_pixel_size
    nx_tomo.instrument.detector.field_of_view = field_of_view
    nx_tomo.instrument.detector.distance = distance
    nx_tomo.instrument.detector.x_flipped = x_flipped
    nx_tomo.instrument.detector.y_flipped = y_flipped
    nx_tomo.energy = energy
    nx_tomo.sample.x_translation = x_translation
    nx_tomo.sample.z_translation = z_translation
    nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION.value] * 12
    nx_tomo.instrument.detector.data = numpy.empty(shape=(12, 10, 10))
    nx_tomo.sample.rotation_angle = numpy.linspace(0, 20, num=12)

    file_path = os.path.join(tmp_path, "nxtomo.nx")
    entry = "entry0000"
    nx_tomo.save(
        file_path=file_path,
        data_path=entry,
    )

    scan = HDF5TomoScan(file_path, entry)

    # 2.0 create the widget and do the edition
    widget = NXtomoEditor()
    widget.setScan(scan=scan)
    # widget.show()

    # 3.0 check data have been corrcetly loaded
    def check_metric(expected_value, current_value):
        if expected_value is None:
            return current_value is None
        return numpy.isclose(expected_value, float(current_value))

    assert check_metric(x_pixel_size, widget._xPixelSizeMetricEntry.getValue())
    assert widget._xPixelSizeMetricEntry._qcbUnit.currentText() == "m"
    assert check_metric(y_pixel_size, widget._yPixelSizeMetricEntry.getValue())
    assert widget._yPixelSizeMetricEntry._qcbUnit.currentText() == "m"

    assert check_metric(distance, widget._distanceMetricEntry.getValue())
    assert widget._distanceMetricEntry._qcbUnit.currentText() == "m"

    assert field_of_view == widget._fieldOfViewCB.currentText()
    assert x_flipped == widget._xFlippedCB.isChecked()
    assert y_flipped == widget._yFlippedCB.isChecked()

    if energy is None:
        assert widget._energyQLE.getValue() is None
    else:
        assert numpy.isclose(energy, widget._energyQLE.getValue())

    def check_translation(expected_value, current_value):
        if expected_value is None:
            return current_value is None
        else:
            u_values = numpy.unique(expected_value)
            if u_values.size == 1:
                return float(current_value) == u_values[0]
            else:
                return current_value is _TranslationMetricEntry.LOADED_ARRAY

    assert check_translation(x_translation, widget._xTranslationQLE.getValue())
    assert widget._xTranslationQLE._qcbUnit.currentText() == "m"
    assert check_translation(z_translation, widget._zTranslationQLE.getValue())
    assert widget._zTranslationQLE._qcbUnit.currentText() == "m"

    # 4.0 edit some parameters
    widget._energyQLE.setText("23.789")
    widget._xPixelSizeMetricEntry.setUnit("nm")
    widget._yPixelSizeMetricEntry.setValue(2.1e-7)
    widget._distanceMetricEntry.setValue("unknown")
    widget._fieldOfViewCB.setCurrentText(_FOV.HALF.value)
    widget._xFlippedCB.setChecked(not x_flipped)
    widget._xTranslationQLE.setValue(1.8)
    widget._xTranslationQLE.setUnit("mm")
    widget._zTranslationQLE.setValue(2.8)
    widget._zTranslationQLE.setUnit("m")

    # 5.0 save
    widget.overwriteNXtomo()

    # 6.0 make sure data have been overwrite
    overwrite_nx_tomo = NXtomo().load(
        file_path=file_path,
        data_path=entry,
    )

    assert overwrite_nx_tomo.energy.value == 23.789
    assert overwrite_nx_tomo.energy.unit == EnergySI.KILOELECTRONVOLT
    if x_pixel_size is None:
        assert overwrite_nx_tomo.instrument.detector.x_pixel_size.value is None
    else:
        assert numpy.isclose(
            overwrite_nx_tomo.instrument.detector.x_pixel_size.value,
            x_pixel_size * MetricSystem.NANOMETER.value,
        )
    assert overwrite_nx_tomo.instrument.detector.y_pixel_size.value == 2.1e-7

    assert overwrite_nx_tomo.instrument.detector.distance.value is None
    assert overwrite_nx_tomo.instrument.detector.field_of_view is _FOV.HALF

    assert overwrite_nx_tomo.instrument.detector.x_flipped is not x_flipped
    assert overwrite_nx_tomo.instrument.detector.y_flipped is y_flipped

    numpy.testing.assert_array_almost_equal(
        overwrite_nx_tomo.sample.x_translation.value,
        numpy.array([1.8 * MetricSystem.MILLIMETER.value] * 12),
    )
    assert overwrite_nx_tomo.sample.x_translation.unit is MetricSystem.METER
    numpy.testing.assert_array_almost_equal(
        overwrite_nx_tomo.sample.z_translation.value,
        numpy.array([2.8 * MetricSystem.METER.value] * 12),
    )
    assert overwrite_nx_tomo.sample.z_translation.unit is MetricSystem.METER
    # end
    widget.setAttribute(qt.Qt.WA_DeleteOnClose)
    widget.close()
    widget = None
